#include <stdio.h>
#include "utils.h"
#include "timer.h"

__global__ void BoxBlurAOSKernel ( uchar4 * d_inputImageRGBA, uchar4 * d_outputImageRGBA, const int imWidth, const int imHeight, const int filterWidth ) {
	int x = threadIdx.x + blockDim.x * blockIdx.x;
	int y = threadIdx.y + blockDim.y * blockIdx.y;

	int imIdx = y * imWidth + x;

	if ( x >= imWidth || y >= imHeight ) {
		return;
	}

	int halfFilterWidth = filterWidth / 2;
	float normalizingFactor = powf( filterWidth, 2 );

	float resultR = 0.0f;
	float resultG = 0.0f;
	float resultB = 0.0f;
	for ( int fx = -halfFilterWidth; fx <= halfFilterWidth; fx++ ) {
		for ( int fy = -halfFilterWidth; fy <= halfFilterWidth; fy++ ) {
		    int image_x = min( max( x + fx, 0 ), ( imWidth - 1 ) );
        	int image_y = min( max( y + fy, 0 ), ( imHeight - 1 ) );

        	int curIdx = image_y * imWidth + image_x;
        	resultR += (d_inputImageRGBA[curIdx].x); 	
        	resultG += (d_inputImageRGBA[curIdx].y); 	
        	resultB += (d_inputImageRGBA[curIdx].z); 		
		}
	}

	resultR /= normalizingFactor;
	resultG /= normalizingFactor;
	resultB /= normalizingFactor;

	d_outputImageRGBA[imIdx] = make_uchar4((unsigned char)resultR, (unsigned char)resultG, (unsigned char)resultB, 255 );
}

__global__ void BoxBlurSOAKernel ( uchar * d_inputChannel, uchar * d_outputChannel, const int imWidth, const int imHeight, const int filterWidth ) {
	int x = threadIdx.x + blockDim.x * blockIdx.x;
	int y = threadIdx.y + blockDim.y * blockIdx.y;

	int imIdx = y * imWidth + x;

	if ( x >= imWidth || y >= imHeight ) {
		return;
	}

	int halfFilterWidth = filterWidth / 2;
	float normalizingFactor = powf( filterWidth, 2 );

	float result = 0.0f;
	for ( int fx = -halfFilterWidth; fx <= halfFilterWidth; fx++ ) {
		for ( int fy = -halfFilterWidth; fy <= halfFilterWidth; fy++ ) {
		    int image_x = min( max( x + fx, 0 ), ( imWidth - 1 ) );
        	int image_y = min( max( y + fy, 0 ), ( imHeight - 1 ) );

        	int curIdx = image_y * imWidth + image_x;
        	result += d_inputChannel[curIdx]; 			
		}
	}

	result /= normalizingFactor;

	d_outputChannel[imIdx] = (uchar)result;
}


__global__ void separateChannels(const uchar4* const inputImageRGBA,
                      unsigned char* const redChannel,
                      unsigned char* const greenChannel,
                      unsigned char* const blueChannel,
                      int imWidth,
                      int imHeight ) {
	int x = threadIdx.x + blockDim.x * blockIdx.x;
	int y = threadIdx.y + blockDim.y * blockIdx.y;

	int imIdx = y * imWidth + x;

	if ( x >= imWidth || y >= imHeight ) {
		return;
	}

    uchar4 rgba = inputImageRGBA[ imIdx ];
    redChannel[ imIdx ] = rgba.x;
    greenChannel[ imIdx ] = rgba.y;
    blueChannel[ imIdx ]  = rgba.z;

}

__global__ void recombineChannels(const unsigned char* const redChannel,
                       const unsigned char* const greenChannel,
                       const unsigned char* const blueChannel,
                       uchar4* const outputImageRGBA,
                       int imWidth,
                       int imHeight ) {
	int x = threadIdx.x + blockDim.x * blockIdx.x;
	int y = threadIdx.y + blockDim.y * blockIdx.y;

	int imIdx = y * imWidth + x;

	if ( x >= imWidth || y >= imHeight ) {
		return;
	}

    unsigned char red   = redChannel[imIdx];
    unsigned char green = greenChannel[imIdx];
    unsigned char blue  = blueChannel[imIdx];

    uchar4 outputPixel = make_uchar4(red, green, blue, 255);

    outputImageRGBA[imIdx] = outputPixel;
}

void BoxBlurAOS ( uchar4 *  d_inputImageRGBA, uchar4 *  d_outputImageRGBA, const int imWidth, const int imHeight ) {
  int blockWidth = 16;
  int filterWidth = 9;

  dim3 blockSize = dim3( blockWidth, blockWidth, 1 );
  dim3 gridSize = dim3( floor(imWidth/blockWidth) + 1, floor(imHeight/blockWidth) + 1, 1 );

  GpuTimer timer;

  timer.Start();
  BoxBlurAOSKernel<<<gridSize,blockSize>>>( d_inputImageRGBA, d_outputImageRGBA, imWidth, imHeight, filterWidth );
  timer.Stop();
  printf("Code ran in: %f msecs.\n", timer.Elapsed());

}

void BoxBlurSOA ( uchar4 *  d_inputImageRGBA, uchar4 *  d_outputImageRGBA, const int imWidth, const int imHeight ) {
  int blockWidth = 16;
  int filterWidth = 9;
  int numPixels = imWidth * imHeight;

  dim3 blockSize = dim3( blockWidth, blockWidth, 1 );
  dim3 gridSize = dim3( floor(imWidth/blockWidth) + 1, floor(imHeight/blockWidth) + 1, 1 );

  unsigned char *d_red, *d_blue, *d_green;
  checkCudaErrors( cudaMalloc((void**)&d_red, numPixels * sizeof( unsigned char ) ) );
  checkCudaErrors( cudaMalloc((void**)&d_green, numPixels * sizeof( unsigned char ) ) );
  checkCudaErrors( cudaMalloc((void**)&d_blue, numPixels * sizeof( unsigned char ) ) );

  unsigned char *d_redBlurred, *d_blueBlurred, *d_greenBlurred;
  checkCudaErrors( cudaMalloc((void**)&d_redBlurred, numPixels * sizeof( unsigned char ) ) );
  checkCudaErrors( cudaMalloc((void**)&d_greenBlurred, numPixels * sizeof( unsigned char ) ) );
  checkCudaErrors( cudaMalloc((void**)&d_blueBlurred, numPixels * sizeof( unsigned char ) ) );
  GpuTimer timer;

  timer.Start();
  separateChannels<<<gridSize,blockSize>>> ( d_inputImageRGBA, d_red, d_green, d_blue, imWidth, imHeight );

  BoxBlurSOAKernel<<<gridSize,blockSize>>>( d_red, d_redBlurred, imWidth, imHeight, filterWidth );
  BoxBlurSOAKernel<<<gridSize,blockSize>>>( d_green, d_greenBlurred, imWidth, imHeight, filterWidth );
  BoxBlurSOAKernel<<<gridSize,blockSize>>>( d_blue, d_blueBlurred, imWidth, imHeight, filterWidth );

  recombineChannels<<<gridSize, blockSize>>> ( d_redBlurred, d_greenBlurred, d_blueBlurred, d_outputImageRGBA, imWidth, imHeight );
  timer.Stop();
  printf("Code ran in: %f msecs.\n", timer.Elapsed());

  checkCudaErrors( cudaFree(d_red) );
  checkCudaErrors( cudaFree(d_green) );
  checkCudaErrors( cudaFree(d_blue) );
  checkCudaErrors( cudaFree(d_redBlurred) );
  checkCudaErrors( cudaFree(d_greenBlurred) );
  checkCudaErrors( cudaFree(d_blueBlurred) );
}

  // unsigned char *d_red, *d_green, *d_blue;
  // unsigned char *d_redBlurred, *d_greenBlurred, *d_blueBlurred;