#pragma once

#include <iomanip>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_runtime_api.h>
#include <cassert>
#include <cmath>
#include <stdio.h>

#define checkCudaErrors(val) check( (val), #val, __FILE__, __LINE__)

typedef unsigned char uchar;

template<typename T>
void check(T err, const char* const func, const char* const file, const int line) {
    if (err != cudaSuccess) {
        fprintf ( stderr, "CUDA error at: %s : %i\n", file, line );
        fprintf ( stderr, "%s %s\n", cudaGetErrorString(err), func );
        exit(1);
    }
}

template<typename T>
void checkResultsExact(const T* const ref, const T* const gpu, size_t numElem) {
    //check that the GPU result matches the CPU result
    for (size_t i = 0; i < numElem; ++i) {
        if (ref[i] != gpu[i]) {
            fprintf ( stderr, "Difference at pos %zu \n", i );
            //the + is magic to convert char to int without messing
            //with other types
            fprintf ( stderr, "Reference %f1.17 \nGPU   :%f1.17", +ref[i], +gpu[i] );
            exit(1);
        }
    }
}

template<typename T>
void checkResultsEps(const T* const ref, const T* const gpu, size_t numElem, double eps1, double eps2) {
    assert ( eps1 >= 0 && eps2 >= 0 );
    unsigned long long totalDiff = 0;
    unsigned numSmallDifferences = 0;
    for (size_t i = 0; i < numElem; ++i) {
        //subtract smaller from larger in case of unsigned types
        T smaller = std::min(ref[i], gpu[i]);
        T larger = std::max(ref[i], gpu[i]);
        T diff = larger - smaller;
        if (diff > 0 && diff <= eps1) {
            numSmallDifferences++;
        }
        else if (diff > eps1) {
            fprintf ( stderr, "Difference at pos %zu exceeds tolerance of %f\n", +i, eps1 );
            fprintf ( stderr, "Reference %1.17f \nGPU   :%1.17f\n", +ref[i], +gpu[i] );
            exit(1);
        }
        totalDiff += diff * diff;
    }
    double percentSmallDifferences = (double)numSmallDifferences / (double)numElem;
    if (percentSmallDifferences > eps2) {
        fprintf(stderr, "Total percentage of non-zero pixel difference between the two images exceeds %f \n", 100.0f * eps2 );
        fprintf(stderr, "Percentage of non-zero pixel differences: %f \n ", 100.0 * percentSmallDifferences );
        exit(1);
    }
}

//Uses the autodesk method of image comparison
//Note the the tolerance here is in PIXELS not a percentage of input pixels
template<typename T>
void checkResultsAutodesk(const T* const ref, const T* const gpu, size_t numElem, double variance, size_t tolerance)
{

    size_t numBadPixels = 0;
    for (size_t i = 0; i < numElem; ++i) {
        T smaller = std::min(ref[i], gpu[i]);
        T larger = std::max(ref[i], gpu[i]);
        T diff = larger - smaller;
        if (diff > variance)
            ++numBadPixels;
    }

    if (numBadPixels > tolerance) {
        fprintf(stderr,"Too many bad pixels in the image. %zu // %zu", numBadPixels, tolerance );
        exit(1);
    }
}

